""" Default parameters for MPE environments """

# Action types
DISCRETE_ACT = "Discrete"
CONTINUOUS_ACT = "Continuous"

# Environment
MAX_STEPS = 25
DAMPING = 0.25
CONTACT_FORCE = 1e2
CONTACT_MARGIN = 1e-3
DT = 0.1

# Colours
AGENT_COLOUR = (115, 243, 115)
ADVERSARY_COLOUR = (243, 115, 115)
OBS_COLOUR = (64, 64, 64)
